from dataclasses import dataclass
from datetime import timedelta
from typing import Generic, ParamSpec, TypeVar

from redis.asyncio import Redis

P = ParamSpec("P")
POther = ParamSpec("POther")
R = TypeVar("R")
ROther = TypeVar("ROther")
WD = TypeVar("WD")


@dataclass
class StreamMessage:
    """
    Dataclass wrapping data stored in the Redis stream.
    """

    message_id: str
    task_id: str
    priority: str


@dataclass
class WrappedContext(Generic[WD]):
    """
    Dataclass wrapping the user-defined context (contained in `deps`)
    with additional information such as try, the Redis connection, and
    other task-specific data.
    """

    deps: WD
    fn_name: str
    redis: Redis
    task_id: str
    timeout: timedelta | int | None
    tries: int
    ttl: timedelta | int | None
    worker_id: str



from typing import Any, AsyncGenerator, Generator

from pytest import fixture
from testcontainers.redis import RedisContainer

from streaq import Worker


@fixture(scope="session")
def redis_container() -> Generator[RedisContainer, Any, None]:
    with RedisContainer() as container:
        yield container


@fixture(scope="function")
def redis_url(redis_container: RedisContainer) -> Generator[str, None, None]:
    yield f"redis://{redis_container.get_container_host_ip()}:{redis_container.port}"
    redis_container.get_client().flushdb()


@fixture(scope="function")
async def worker(redis_url: str) -> AsyncGenerator[Worker, None]:
    w = Worker(
        redis_url=redis_url,
        queue_name="test",
        handle_signals=False,
        with_scheduler=True,
    )
    async with w:
        yield w
    await w.close()



import logging.config
import os
import sys
from multiprocessing import Process
from typing import Annotated, cast

from typer import Exit, Option, Typer
from watchfiles import run_process

from streaq import VERSION, logger
from streaq.utils import default_log_config, import_string
from streaq.worker import Worker

cli = Typer()


def version_callback(value: bool) -> None:
    if value:
        print(f"streaQ v{VERSION}")
        raise Exit()


@cli.command()
def main(
    worker_path: str,
    workers: Annotated[
        int, Option("--workers", "-w", help="Number of worker processes to spin up")
    ] = 1,
    burst: Annotated[
        bool,
        Option(
            "--burst", "-b", help="Whether to shut down worker when the queue is empty"
        ),
    ] = False,
    reload: Annotated[
        bool,
        Option(
            "--reload", "-r", help="Whether to reload the worker upon changes detected"
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Option(
            "--verbose",
            "-v",
            help="Whether to use logging.DEBUG instead of logging.INFO",
        ),
    ] = False,
    version: Annotated[
        bool,
        Option("--version", callback=version_callback, help="Show installed version"),
    ] = False,
):
    processes = []
    if workers > 1:
        processes.extend(
            [
                Process(
                    target=run_worker,
                    args=(worker_path, burst, reload, verbose, False),
                )
                for _ in range(workers - 1)
            ]
        )
    for p in processes:
        p.start()
    # only one runs a scheduler if with_scheduler is None
    run_worker(worker_path, burst, reload, verbose, True)
    for p in processes:
        p.join()


def run_worker(path: str, burst: bool, watch: bool, verbose: bool, schedule: bool):
    """
    Run a worker with the given options.
    """
    if watch:
        run_process(
            ".",
            target=_run_worker,
            args=(path, burst, verbose, schedule),
            callback=lambda _: logger.info("changes detected, reloading"),
        )
    else:
        _run_worker(path, burst, verbose, schedule)


def _run_worker(path: str, burst: bool, verbose: bool, schedule: bool):
    sys.path.append(os.getcwd())
    logging.config.dictConfig(default_log_config(verbose))
    worker = cast(Worker, import_string(path))
    worker.burst = burst
    if worker.with_scheduler is None:
        worker.with_scheduler = schedule
    worker.run_sync()



import asyncio
import hashlib
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Concatenate,
    Coroutine,
    Generic,
)
from uuid import UUID, uuid4

from crontab import CronTab
from redis.asyncio import Redis
from redis.typing import EncodableT

from streaq import logger
from streaq.constants import (
    DEFAULT_TTL,
    REDIS_DEPENDENCIES,
    REDIS_DEPENDENTS,
    REDIS_MESSAGE,
    REDIS_PREFIX,
    REDIS_RESULT,
    REDIS_TASK,
)
from streaq.types import P, POther, R, ROther, WD, WrappedContext
from streaq.utils import StreaqError, datetime_ms, now_ms, to_seconds, to_ms

if TYPE_CHECKING:
    from streaq.worker import Worker


class StreaqRetry(StreaqError):
    """
    An exception you can manually raise in your tasks to make sure the task
    is retried.

    :param msg: error message to show
    :param delay:
        amount of time to wait before retrying the task; if none, will be the
        square of the number of attempts in seconds
    """

    def __init__(self, msg: str, delay: timedelta | int | None = None):
        super().__init__(msg)
        self.delay = delay


class TaskPriority(str, Enum):
    """
    Enum of possible task priority levels.
    """

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class TaskStatus(str, Enum):
    """
    Enum of possible task statuses.
    """

    PENDING = "pending"
    QUEUED = "queued"
    RUNNING = "running"
    SCHEDULED = "scheduled"
    DONE = "done"


@dataclass
class TaskData:
    """
    Dataclass containing additional task information not stored locally,
    such as try and time enqueued.
    """

    fn_name: str
    enqueue_time: int
    task_try: int | None = None
    scheduled: datetime | None = None


@dataclass
class TaskResult(Generic[R]):
    """
    Dataclass wrapping the result of a task with additional information
    like run time and whether execution terminated successfully.
    """

    success: bool
    result: R | Exception
    start_time: int
    finish_time: int
    queue_name: str


@dataclass
class Task(Generic[R]):
    """
    Represents a task that has been enqueued or scheduled.
    """

    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    parent: "RegisteredCron | RegisteredTask"
    _after: "Task | None" = None
    _triggers: "Task | None" = None
    id: str = ""

    def __post_init__(self):
        if self.parent.unique:
            deterministic_hash = hashlib.sha256(
                self.parent.fn_name.encode()
            ).hexdigest()
            self.id = UUID(bytes=bytes.fromhex(deterministic_hash[:32]), version=4).hex
        else:
            self.id = uuid4().hex

    async def start(
        self,
        after: str | list[str] | None = None,
        delay: timedelta | int | None = None,
        schedule: datetime | None = None,
        priority: TaskPriority = TaskPriority.LOW,
    ) -> "Task[R]":
        """
        Enqueues a task immediately, for running after a delay, or for running
        at a specified time.

        :param after: task ID(s) to wait for before running this task
        :param delay: duration to wait before running the task
        :param schedule: datetime at which to run the task
        :param priority: priority queue to insert the task

        :return: self
        """
        after = self._after.id if self._after else after
        if (delay and schedule) or (delay and after) or (schedule and after):
            raise StreaqError(
                "Use one of 'delay', 'schedule', or 'after' when enqueuing tasks, not "
                "multiple!"
            )
        if not self.parent.worker.scripts:
            raise StreaqError(
                "Worker did not initialize correctly, are you using the async context "
                "manager?"
            )
        enqueue_time = now_ms()
        if schedule:
            score = datetime_ms(schedule)
        elif delay is not None:
            score = enqueue_time + to_ms(delay)
        else:
            score = None
        ttl = (score or enqueue_time) - enqueue_time + DEFAULT_TTL
        data = self.serialize(enqueue_time)
        run_now = False
        if after:
            if isinstance(after, str):
                after = [after]
            run_now = await self.parent.worker.scripts["add_dependencies"](
                keys=[
                    self._task_key(REDIS_TASK),
                    self.id,
                    REDIS_PREFIX + self.queue + REDIS_DEPENDENTS,
                    REDIS_PREFIX + self.queue + REDIS_DEPENDENCIES,
                    REDIS_PREFIX + self.queue + REDIS_RESULT,
                ],
                args=[data, ttl] + after,
            )
        if not after or run_now:
            keys = [
                self.parent.worker._stream_key,
                self._task_key(REDIS_MESSAGE),
                self._task_key(REDIS_TASK),
                self.parent.worker._queue_key,
            ]
            args = [self.id, ttl, data, priority]
            if score:
                args.append(score)
            res = await self.parent.worker.scripts["publish_task"](keys=keys, args=args)
            if res == 0:
                logger.debug("Task is unique and already exists, not enqueuing!")
                return self

        return self

    def then(
        self, task: "RegisteredTask[WD, POther, ROther]", **kwargs
    ) -> "Task[ROther]":
        """
        Enqueues the given task as a dependent of this one. Positional arguments will
        come from the previous task's output (tuple outputs will be unpacked), and any
        additional arguments can be passed as kwargs.

        :param task: task to feed output to

        :return: task object for newly created, dependent task
        """
        self._triggers = Task((), kwargs, task)
        self._triggers._after = self
        return self._triggers

    async def _chain(self) -> "Task[R]":
        # traverse backwards
        if self._after:
            await self._after
        return await self.start()

    def __await__(self):
        return self._chain().__await__()

    def _task_key(self, mid: str) -> str:
        return REDIS_PREFIX + self.queue + mid + self.id

    def serialize(self, enqueue_time: int) -> EncodableT:
        """
        Serializes the task data for sending to the queue.

        :param enqueue_time: the time at which the task was enqueued

        :return: serialized task data
        """
        try:
            data = {
                "f": self.parent.fn_name,
                "a": self.args,
                "k": self.kwargs,
                "t": enqueue_time,
            }
            if self._after:
                data["A"] = self._after.id
            if self._triggers:
                data["T"] = self._triggers.id
            return self.parent.worker.serializer(data)
        except Exception as e:
            raise StreaqError(f"Unable to serialize task {self.parent.fn_name}:") from e

    async def status(self) -> TaskStatus:
        """
        Fetch the current status of the task.

        :return: current task status
        """
        return await self.parent.worker.status_by_id(self.id)

    async def result(self, timeout: timedelta | int | None = None) -> TaskResult[R]:
        """
        Wait for and return the task's result, optionally with a timeout.

        :param timeout: amount of time to wait before raising a `TimeoutError`

        :return: wrapped result object
        """
        return await self.parent.worker.result_by_id(self.id, timeout=timeout)

    async def abort(self, timeout: timedelta | int = 5) -> bool:
        """
        Notify workers that the task should be aborted.

        :param timeout: how long to wait to confirm abortion was successful

        :return: whether the task was aborted successfully
        """
        return await self.parent.worker.abort_by_id(self.id, timeout=timeout)

    async def info(self) -> TaskData:
        """
        Fetch info about a previously enqueued task.

        :return: task info object
        """
        return await self.parent.worker.info_by_id(self.id)

    @property
    def redis(self) -> Redis:
        return self.parent.worker.redis

    @property
    def queue(self) -> str:
        return self.parent.worker.queue_name


@dataclass
class RegisteredTask(Generic[WD, P, R]):
    fn: Callable[Concatenate[WrappedContext[WD], P], Coroutine[Any, Any, R]]
    max_tries: int | None
    timeout: timedelta | int | None
    ttl: timedelta | int | None
    unique: bool
    worker: "Worker"
    _fn_name: str | None = None

    def enqueue(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Task[R]:
        """
        Serialize the task and send it to the queue for later execution by an active worker.
        Though this isn't async, it should be awaited as it returns an object that should be.
        """
        return Task(args, kwargs, self)

    async def run(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """
        Run the task in the local event loop with the given params and return the result.
        This skips enqueuing and result storing in Redis.
        """
        deps = WrappedContext(
            deps=self.worker.deps,
            fn_name=self.fn_name,
            redis=self.worker.redis,
            task_id=uuid4().hex,
            timeout=self.timeout,
            tries=1,
            ttl=self.ttl,
            worker_id=self.worker.id,
        )
        try:
            return await asyncio.wait_for(
                self.fn(deps, *args, **kwargs),
                to_seconds(self.timeout) if self.timeout else None,
            )
        except asyncio.TimeoutError as e:
            raise e

    @property
    def fn_name(self) -> str:
        return self._fn_name or self.fn.__qualname__

    def __repr__(self):
        return f"<Task fn={self.fn_name} timeout={self.timeout} ttl={self.ttl}>"


@dataclass
class RegisteredCron(Generic[WD, R]):
    fn: Callable[[WrappedContext[WD]], Coroutine[Any, Any, R]]
    max_tries: int | None
    crontab: CronTab
    timeout: timedelta | int | None
    ttl: timedelta | int | None
    unique: bool
    worker: "Worker"

    def enqueue(self) -> Task[R]:
        """
        Serialize the task and send it to the queue for later execution by an active worker.
        Though this isn't async, it should be awaited as it returns an object that should be.
        """
        return Task((), {}, self)

    async def run(self) -> R:
        """
        Run the task in the local event loop and return the result.
        This skips enqueuing and result storing in Redis.
        """
        deps = WrappedContext(
            deps=self.worker.deps,
            fn_name=self.fn_name,
            redis=self.worker.redis,
            task_id=uuid4().hex,
            timeout=self.timeout,
            tries=1,
            ttl=self.ttl,
            worker_id=self.worker.id,
        )
        try:
            return await asyncio.wait_for(
                self.fn(deps),
                to_seconds(self.timeout) if self.timeout else None,
            )
        except asyncio.TimeoutError as e:
            raise e

    @property
    def fn_name(self) -> str:
        return self.fn.__qualname__

    def schedule(self) -> datetime:
        """
        Datetime of next run.
        """
        return datetime.fromtimestamp(self.next() / 1000, tz=self.worker.tz)

    def next(self) -> int:
        """
        Timestamp in milliseconds of next run.
        """
        return round((time() + self.delay) * 1000)

    @property
    def delay(self) -> float:
        return self.crontab.next(now=datetime.now(self.worker.tz))  # type: ignore

    def __repr__(self):
        return (
            f"<Cron fn={self.fn_name} timeout={self.timeout} "
            f"schedule={self.schedule()}>"
        )



from redis.asyncio import Redis
from redis.commands.core import AsyncScript


ADD_DEPENDENCIES = """
local task_key = KEYS[1]
local task_id = KEYS[2]
local dependents_key = KEYS[3]
local dependencies_key = KEYS[4]
local prefix = KEYS[5]


local task_data = ARGV[1]
local ttl = ARGV[2]

if redis.call('set', task_key, task_data, 'nx', 'px', ttl) == nil then
  return
end

local modified = 0
for i=3, #ARGV do
  local dep_id = ARGV[i]
  if redis.call('exists', prefix .. dep_id) ~= 1 then
    modified = modified + 1
    redis.call('sadd', dependencies_key .. task_id, dep_id)
    redis.call('sadd', dependents_key .. dep_id, task_id)
  end
end

if modified == 0 then
  return 1
end

return
"""

PUBLISH_TASK = """
local stream_key = KEYS[1]
local message_key = KEYS[2]
local task_key = KEYS[3]
local queue_key = KEYS[4]

local task_id = ARGV[1]
local ttl = ARGV[2]
local task_data = ARGV[3]
local priority = ARGV[4]
local score = ARGV[5]

if not redis.call('set', task_key, task_data, 'nx', 'px', ttl) then
  return 0
end

if not score then
  local message_id = redis.call('xadd', stream_key .. priority, '*', 'task_id', task_id)
  redis.call('set', message_key, message_id, 'px', ttl)
  return message_id
else
  redis.call('zadd', queue_key, score, task_id)
  return 1
end
"""

PUBLISH_DELAYED_TASK = """
local delayed_queue_key = KEYS[1]
local stream_key = KEYS[2]
local task_message_id_key = KEYS[3]

local task_id = ARGV[1]
local task_message_id_expire_ms = ARGV[2]
local priority = ARGV[3]

local score = redis.call('zscore', delayed_queue_key, task_id)
if score == nil or score == false then
  return 0
end

local message_id = redis.call('xadd', stream_key .. priority, '*', 'task_id', task_id)
redis.call('set', task_message_id_key, message_id, 'px', task_message_id_expire_ms)
redis.call('zrem', delayed_queue_key, task_id)
return 1
"""

RETRY_TASK = """
local stream_key = KEYS[1]
local message_key = KEYS[2]

local task_id = ARGV[1]
local expire_ms = ARGV[2]

local message_id = redis.call('xadd', stream_key, '*', 'task_id', task_id)
redis.call('set', message_key, message_id, 'px', expire_ms)
return message_id
"""

PUBLISH_DEPENDENT = """
local stream_key = KEYS[1]
local message_key = KEYS[2]
local dep_id = KEYS[3]

local ttl = ARGV[1]

local message_id = redis.call('xadd', stream_key, '*', 'task_id', dep_id)
redis.call('set', message_key, message_id, 'px', ttl)
"""

UNPUBLISH_DEPENDENT = """
local task_key = KEYS[1]
local result_key = KEYS[2]
local channel_key = KEYS[3]
local dep_id = KEYS[4]

local result_data = ARGV[1]

redis.call('del', task_key)
redis.call('set', result_key, result_data, 'ex', 300)
redis.call('publish', channel_key, dep_id)
"""

FAIL_DEPENDENTS = """
local dependents_key = KEYS[1]
local dependencies_key = KEYS[2]
local task_id = KEYS[3]

local function traverse(task_id, failed)
  if not failed[task_id] then
    failed[task_id] = true
    for _, dep_id in ipairs(redis.call('smembers', dependents_key .. task_id)) do
      traverse(dep_id, failed)
      redis.call('srem', dependencies_key .. dep_id, task_id)
    end
  redis.call('del', dependents_key .. task_id)
  end
end

local visited = {}
traverse(task_id, visited)
local failed = {}
for tid, _ in pairs(visited) do
  if tid ~= task_id then
    table.insert(failed, tid)
  end
end

return failed
"""

UPDATE_DEPENDENTS = """
local dependents_key = KEYS[1]
local dependencies_key = KEYS[2]
local task_id = KEYS[3]

local runnable = {}

local deps = redis.call('smembers', dependents_key .. task_id)
for i = 1, #deps do
  redis.call('srem', dependencies_key .. deps[i], task_id)
  if #redis.call('smembers', dependencies_key .. deps[i]) == 0 then
    table.insert(runnable, deps[i])
  end
end

redis.call('del', dependents_key .. task_id, dependencies_key .. task_id)

return runnable
"""

UNCLAIM_IDLE_TASKS = """
local sorted_set_key = KEYS[1]
local stream_key = KEYS[2]
local group_name = KEYS[3]
local consumer_name = KEYS[4]
local message_key = KEYS[5]

local current_time = ARGV[1]
local ttl = ARGV[2]

local timed_out = redis.call('zrangebyscore', sorted_set_key, '-inf', current_time)
if #timed_out > 0 then
  local claimed = redis.call('xclaim', stream_key, group_name, consumer_name, 0, unpack(timed_out))
  for _, message in ipairs(claimed) do
    redis.call('xack', stream_key, group_name, message[1])
    redis.call('xdel', stream_key, message[1])
    local message_id = redis.call('xadd', stream_key, '*', unpack(message[2]))
    redis.call('set', message_key .. message[2][2], message_id, 'px', ttl)
  end
  redis.call('zrem', sorted_set_key, unpack(timed_out))
end

return timed_out
"""


def register_scripts(redis: Redis) -> dict[str, AsyncScript]:
    return {
        "add_dependencies": redis.register_script(ADD_DEPENDENCIES),
        "publish_task": redis.register_script(PUBLISH_TASK),
        "publish_delayed_task": redis.register_script(PUBLISH_DELAYED_TASK),
        "retry_task": redis.register_script(RETRY_TASK),
        "fail_dependents": redis.register_script(FAIL_DEPENDENTS),
        "publish_dependent": redis.register_script(PUBLISH_DEPENDENT),
        "unpublish_dependent": redis.register_script(UNPUBLISH_DEPENDENT),
        "update_dependents": redis.register_script(UPDATE_DEPENDENTS),
        "unclaim_idle_tasks": redis.register_script(UNCLAIM_IDLE_TASKS),
    }



import asyncio
import time
from datetime import datetime, timedelta
from typing import Callable, Coroutine

import pytest

from streaq import StreaqError, WrappedContext, Worker
from streaq.task import StreaqRetry, TaskPriority, TaskStatus


async def test_result_timeout(worker: Worker):
    @worker.task(ttl=0)
    async def foobar(ctx: WrappedContext[None]) -> bool:
        return False

    task = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    with pytest.raises(StreaqError):
        await task.result(3)


async def test_task_timeout(worker: Worker):
    @worker.task(timeout=1)
    async def foobar(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(5)

    task = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    res = await task.result(3)
    assert not res.success
    assert isinstance(res.result, asyncio.TimeoutError)


async def test_task_status(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(1)

    task = foobar.enqueue()
    assert await task.status() == TaskStatus.PENDING
    await task.start()
    task2 = await foobar.enqueue().start(delay=5)
    assert await task.status() == TaskStatus.QUEUED
    worker.loop.create_task(worker.run_async())
    await task.result(3)
    assert await task.status() == TaskStatus.DONE
    assert await task2.status() == TaskStatus.SCHEDULED


async def test_task_status_running(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(3)

    task = await foobar.enqueue().start()
    worker.loop.create_task(worker.run_async())
    await asyncio.sleep(1)
    assert await task.status() == TaskStatus.RUNNING
    assert await task.abort()


async def test_task_cron(worker: Worker):
    @worker.cron("30 9 1 1 *")
    async def cron1(ctx: WrappedContext[None]) -> None:
        pass

    @worker.cron("* * * * * * *")  # once/second
    async def cron2(ctx: WrappedContext[None]) -> bool:
        return True

    schedule = cron1.schedule()
    assert schedule.day == 1 and schedule.month == 1
    await cron1.run()
    task = cron2.enqueue()  # by not awaiting we just get the task obj
    worker.loop.create_task(worker.run_async())
    res = await task.result(3)
    assert res.result and res.success


async def test_task_info(redis_url: str):
    worker = Worker(redis_url=redis_url)

    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        pass

    async with worker:
        task = await foobar.enqueue().start(delay=5)
        task2 = await foobar.enqueue()
        assert (await task.info()).scheduled is not None
        assert (await task2.info()).scheduled is None


async def test_task_retry(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> int:
        if ctx.tries < 3:
            raise StreaqRetry("Retrying!")
        return ctx.tries

    task = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    res = await task.result(7)
    assert res.success
    assert res.result == 3


async def test_task_retry_with_delay(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> int:
        if ctx.tries == 1:
            raise StreaqRetry("Retrying!", delay=3)
        return ctx.tries

    task = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    with pytest.raises(asyncio.TimeoutError):
        await task.result(3)
    res = await task.result(1)
    assert res is not None
    assert res.success
    assert res.result == 2


async def test_task_failure(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        raise Exception("That wasn't supposed to happen!")

    task = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    res = await task.result(3)
    assert not res.success
    assert isinstance(res.result, Exception)


async def test_task_retry_no_delay(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> bool:
        if ctx.tries == 1:
            raise StreaqRetry("Retrying!", delay=0)
        return True

    task = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    res = await task.result(3)
    assert res is not None
    assert res.success
    assert res.result


async def test_task_max_retries(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        raise StreaqRetry("Retrying!", delay=0)

    task = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    res = await task.result(3)
    assert res is not None
    assert not res.success
    assert isinstance(res.result, StreaqError)


async def test_task_failed_abort(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> bool:
        return True

    worker.burst = True
    task = await foobar.enqueue()
    await worker.run_async()
    result = await task.result(3)
    assert result.success
    assert result.result
    aborted = await task.abort()
    assert not aborted


async def test_task_nonexistent_or_finished_dependency(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        pass

    task = await foobar.enqueue().start(after="nonexistent")
    worker.loop.create_task(worker.run_async())
    with pytest.raises(asyncio.TimeoutError):
        await task.result(3)


async def test_task_dependency(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(1)

    task = await foobar.enqueue().start(delay=1)
    task2 = await foobar.enqueue().start(after=task.id)
    worker.loop.create_task(worker.run_async())
    assert await task2.status() == TaskStatus.PENDING
    await task.result(3)
    result = await task2.result(3)
    assert result.success


async def test_task_dependency_multiple(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(1)

    task = await foobar.enqueue().start()
    task2 = await foobar.enqueue().start(after=task.id)
    task3 = await foobar.enqueue().start(after=[task.id, task2.id])
    worker.loop.create_task(worker.run_async())
    assert await task2.status() == TaskStatus.PENDING
    assert await task3.status() == TaskStatus.PENDING
    res1 = await task.result(3)
    assert res1.success
    assert await task3.status() == TaskStatus.PENDING
    res2 = await task2.result(3)
    assert res2.success
    res3 = await task3.result(3)
    assert res3.success


async def test_task_dependency_failed(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        raise Exception("Oh no!")

    @worker.task()
    async def do_nothing(ctx: WrappedContext[None]) -> None:
        pass

    task = await foobar.enqueue().start()
    dep = await do_nothing.enqueue().start(after=task.id)
    worker.loop.create_task(worker.run_async())
    res = await dep.result(3)
    assert not res.success
    assert isinstance(res.result, StreaqError)


async def test_sync_task(worker: Worker):
    @worker.task()
    def foobar(ctx: WrappedContext[None]) -> None:
        time.sleep(2)

    task = await foobar.enqueue()
    task2 = await foobar.enqueue()
    worker.loop.create_task(worker.run_async())
    # this would time out if these were running sequentially
    results = await asyncio.gather(task.result(3), task2.result(3))
    assert all(res.success for res in results)


async def test_unsafe_enqueue(redis_url: str, worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None], ret: int) -> int:
        return ret

    worker.loop.create_task(worker.run_async())
    worker2 = Worker(redis_url=redis_url, queue_name="test")

    async with worker2:
        task = await worker2.enqueue_unsafe(foobar.fn_name, 42)
        res = await task.result(3)
        assert res.success
        assert res.result == 42


async def test_chained_failed_dependencies(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        raise Exception("Oh no!")

    @worker.task()
    async def child(ctx: WrappedContext[None]) -> None:
        pass

    task = await foobar.enqueue().start(delay=3)
    dep1 = await child.enqueue().start(after=task.id)
    dep2 = await child.enqueue().start(after=[task.id, dep1.id])
    worker.loop.create_task(worker.run_async())
    await asyncio.sleep(1)
    assert await task.abort(3)
    res1 = await dep1.result(3)
    res2 = await dep2.result(3)
    assert not res1.success and isinstance(res1.result, StreaqError)
    assert not res2.success and isinstance(res2.result, StreaqError)


async def test_task_priorities(redis_url: str):
    worker = Worker(redis_url=redis_url, queue_name="test", concurrency=4)

    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(1)

    async with worker:
        low = [foobar.enqueue() for _ in range(4)]
        high = [foobar.enqueue() for _ in range(4)]
        await asyncio.gather(  # enqueue all tasks
            *[t.start(priority=TaskPriority.LOW) for t in low],
            *[t.start(priority=TaskPriority.HIGH) for t in high],
        )
        worker.loop.create_task(worker.run_async())
        results = await asyncio.gather(*[t.result(3) for t in high])
        statuses = await asyncio.gather(*[t.status() for t in low])
        assert all(res.success for res in results)
        assert all(status != TaskStatus.DONE for status in statuses)
    await worker.close()


async def test_scheduled_task(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        pass

    dt = datetime.now() + timedelta(seconds=1)
    task = await foobar.enqueue().start(schedule=dt)
    worker.loop.create_task(worker.run_async())
    assert await task.status() == TaskStatus.SCHEDULED
    res = await task.result(3)
    assert res.success


async def test_bad_start_params(redis_url: str):
    worker = Worker(redis_url=redis_url)

    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        pass

    with pytest.raises(StreaqError):
        await foobar.enqueue().start(delay=1, schedule=datetime.now())
    with pytest.raises(StreaqError):
        await foobar.enqueue().start(delay=1, after="foobar")
    with pytest.raises(StreaqError):
        await foobar.enqueue().start(schedule=datetime.now(), after="foobar")


async def test_enqueue_unique_task(redis_url: str):
    worker = Worker(redis_url=redis_url)

    @worker.task(unique=True)
    async def foobar(ctx: WrappedContext[None]) -> None:
        pass

    async with worker:
        task = await foobar.enqueue()
        task2 = await foobar.enqueue()
        assert task.id == task2.id
        res = await asyncio.gather(task.info(), task2.info())
        assert res[0] == res[1]


async def test_failed_abort(worker: Worker):
    @worker.task(ttl=0)
    async def foobar(ctx: WrappedContext[None]) -> None:
        pass

    task = await foobar.enqueue().start()
    worker.loop.create_task(worker.run_async())
    await asyncio.sleep(1)
    assert not await task.abort(1)


async def test_cron_run(redis_url: str):
    worker = Worker(redis_url=redis_url)

    @worker.cron("* * * * * * *")
    async def cron1(ctx: WrappedContext[None]) -> bool:
        return True

    @worker.cron("* * * * * * *", timeout=1)
    async def cron2(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(3)

    async with worker:
        assert await cron1.run()
        with pytest.raises(asyncio.TimeoutError):
            await cron2.run()


async def test_sync_cron(worker: Worker):
    @worker.cron("* * * * * * *")
    def cronjob(ctx: WrappedContext[None]) -> None:
        time.sleep(1)

    worker.loop.create_task(worker.run_async())
    res = await cronjob.enqueue().result(3)
    assert res.success and res.result is None


async def test_cron_multiple_runs(worker: Worker):
    key = "runs"

    @worker.cron("* * * * * * *")
    async def cronjob(ctx: WrappedContext[None]) -> None:
        await ctx.redis.incr(key)

    worker.loop.create_task(worker.run_async())
    await asyncio.sleep(3)
    runs = await worker.redis.get(key)
    assert int(runs) > 1


async def test_middleware(worker: Worker):
    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> int:
        return 2

    @worker.middleware
    def double(ctx: WrappedContext[None], task: Callable[..., Coroutine]):
        async def wrapper(*args, **kwargs):
            result = await task(*args, **kwargs)
            return result * 2

        return wrapper

    worker.loop.create_task(worker.run_async())
    task = await foobar.enqueue()
    res = await task.result(3)
    assert res.success
    assert res.result == 4


async def test_task_pipeline(worker: Worker):
    @worker.task()
    async def double(ctx: WrappedContext[None], val: int) -> int:
        return val * 2

    @worker.task()
    async def is_even(ctx: WrappedContext[None], val: int) -> bool:
        return val % 2 == 0

    worker.loop.create_task(worker.run_async())
    async with worker:
        task = await double.enqueue(2).then(double).then(is_even)
        res = await task.result(3)
        assert res.result and res.success



import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import AsyncIterator

from httpx import AsyncClient
from streaq import Worker, WrappedContext


@dataclass
class Context:
    """
    Type safe way of defining the dependencies of your tasks.
    e.g. HTTP client, database connection, settings.
    """

    http_client: AsyncClient


@asynccontextmanager
async def lifespan(worker: Worker) -> AsyncIterator[Context]:
    async with AsyncClient() as http_client:
        yield Context(http_client)


worker = Worker(redis_url="redis://localhost:6379", lifespan=lifespan)


@worker.task(timeout=5)
async def fetch(ctx: WrappedContext[Context], url: str) -> int:
    # ctx.deps here is of type Context, enforced by static typing
    # ctx also provides access the Redis connection, retry count, etc.
    r = await ctx.deps.http_client.get(url)
    return len(r.text)


@worker.cron("* * * * mon-fri")
async def cronjob(ctx: WrappedContext[Context]) -> None:
    print("It's a bird... It's a plane... It's CRON!")


async def main():
    async with worker:
        await fetch.enqueue("https://tastyware.dev/")
        # this will be run directly locally, not enqueued
        await fetch.run("https://github.com/python-arq/arq")
        # enqueue returns a task object that can be used to get results/info
        task = await fetch.enqueue("https://github.com/tastyware/streaq").start(delay=3)
        print(await task.info())
        print(await task.result(timeout=5))


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
from signal import Signals
from typing import AsyncIterator
from uuid import uuid4

import pytest

from streaq.task import TaskStatus
from streaq.types import WrappedContext
from streaq.utils import StreaqError
from streaq.worker import Worker

NAME_STR = "Freddy"


async def test_worker_redis(redis_url: str):
    worker = Worker(redis_url=redis_url)
    await worker.redis.ping()


@dataclass
class Context:
    name: str


@asynccontextmanager
async def deps(worker: Worker) -> AsyncIterator[Context]:
    yield Context(NAME_STR)


async def test_lifespan(redis_url: str):
    worker = Worker(redis_url=redis_url, lifespan=deps)

    @worker.task()
    async def foobar(ctx: WrappedContext[Context]) -> str:
        return ctx.deps.name

    @worker.task(timeout=1)
    async def foobar2(ctx: WrappedContext[Context]) -> None:
        await asyncio.sleep(3)

    async with worker:
        res = await foobar.run()
        assert res == NAME_STR
        with pytest.raises(asyncio.TimeoutError):
            await foobar2.run()


async def test_health_check(redis_url: str):
    worker = Worker(
        redis_url=redis_url,
        health_crontab="* * * * * * *",
        queue_name="test",
        handle_signals=False,
        with_scheduler=True,
    )
    await worker.redis.flushdb()
    worker.loop.create_task(worker.run_async())
    await asyncio.sleep(2)
    worker_health = await worker.redis.hget(worker._health_key, worker.id)  # type: ignore
    redis_health = await worker.redis.hget(worker._health_key, "redis")  # type: ignore
    assert worker_health is not None
    assert redis_health is not None
    await worker.close()


async def test_queue_size(redis_url: str):
    worker = Worker(queue_name=uuid4().hex, redis_url=redis_url)
    assert await worker.queue_size() == 0


def raise_error(*arg, **kwargs) -> None:
    raise Exception("Couldn't serialize/deserialize!")


async def test_bad_serializer(redis_url: str):
    worker = Worker(redis_url=redis_url, serializer=raise_error)  # type: ignore

    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        print("This can't print!")

    async with worker:
        with pytest.raises(StreaqError):
            await foobar.enqueue()


async def test_bad_deserializer(redis_url: str):
    worker = Worker(redis_url=redis_url, deserializer=raise_error)

    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        print("This can't print!")

    worker.burst = True
    async with worker:
        task = await foobar.enqueue()

    await worker.run_async()
    with pytest.raises(StreaqError):
        await task.result(3)


async def test_uninitialized_worker(redis_url: str):
    worker = Worker(redis_url=redis_url)

    @worker.task()
    async def foobar(ctx: WrappedContext[None]) -> None:
        pass

    with pytest.raises(StreaqError):
        await foobar.run()
    with pytest.raises(StreaqError):
        await foobar.enqueue()


async def test_active_tasks(worker: Worker):
    @worker.task()
    async def foo(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(3)

    n_tasks = 5
    for _ in range(n_tasks):
        await foo.enqueue()
    worker.loop.create_task(worker.run_async())
    await asyncio.sleep(1)
    assert worker.active == n_tasks


async def test_handle_signal(worker: Worker):
    @worker.task()
    async def foo(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(3)

    worker._handle_signals = True
    worker.loop.create_task(worker.run_async())
    await asyncio.sleep(1)
    worker.handle_signal(Signals.SIGINT)
    await asyncio.sleep(1)
    task = await foo.enqueue()
    assert await task.status() == TaskStatus.QUEUED


async def test_reclaim_idle_task(redis_url: str):
    async def foo(ctx: WrappedContext[None]) -> None:
        await asyncio.sleep(2)

    worker1 = Worker(
        redis_url=redis_url,
        queue_name="test",
        handle_signals=False,
    )
    foo1 = worker1.task(timeout=3)(foo)
    async with worker1:
        task = await foo1.enqueue()
        await asyncio.wait_for(worker1.run_async(), 1)
        # simulate abrupt shutdown
        for t in worker1.task_wrappers.values():
            t.cancel()
        for t in worker1.tasks.values():
            t.cancel()
        await asyncio.sleep(2)
        assert await task.status() == TaskStatus.RUNNING

    worker2 = Worker(
        redis_url=redis_url,
        queue_name="test",
        handle_signals=False,
        with_scheduler=True,
    )
    worker2.task(timeout=3)(foo)
    worker2.loop.create_task(worker2.run_async())
    assert (await task.result(5)).success
    await worker2.close()



import asyncio
from functools import partial
import pickle
import signal
from collections import defaultdict
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from datetime import datetime, timedelta, timezone, tzinfo
from signal import Signals
from time import time
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Concatenate,
    Coroutine,
    Generic,
    cast,
)
from uuid import uuid4

from anyio.abc import CapacityLimiter
from crontab import CronTab
from redis.asyncio import Redis
from redis.asyncio.client import PubSub
from redis.asyncio.sentinel import Sentinel
from redis.commands.core import AsyncScript
from redis.typing import EncodableT

from streaq import logger
from streaq.constants import (
    DEFAULT_QUEUE_NAME,
    DEFAULT_TTL,
    REDIS_ABORT,
    REDIS_CHANNEL,
    REDIS_DEPENDENCIES,
    REDIS_DEPENDENTS,
    REDIS_GROUP,
    REDIS_HEALTH,
    REDIS_MESSAGE,
    REDIS_PREFIX,
    REDIS_QUEUE,
    REDIS_RESULT,
    REDIS_RETRY,
    REDIS_RUNNING,
    REDIS_STREAM,
    REDIS_TASK,
    REDIS_TIMEOUT,
)
from streaq.lua import register_scripts
from streaq.types import P, R, WD, StreamMessage, WrappedContext
from streaq.utils import StreaqError, asyncify, now_ms, to_ms, to_seconds, to_tuple
from streaq.task import (
    RegisteredCron,
    RegisteredTask,
    StreaqRetry,
    Task,
    TaskData,
    TaskPriority,
    TaskResult,
    TaskStatus,
)

"""
Empty object representing uninitialized dependencies. This is distinct from
None because it is possible for an initialized worker to have no dependencies
(implying the WD type is 'None').
"""
uninitialized = object()


@asynccontextmanager
async def _lifespan(worker: "Worker") -> AsyncIterator[None]:
    yield


async def _placeholder(self, *args): ...


class Worker(Generic[WD]):
    """
    Worker object that fetches and executes tasks from a queue.

    :param redis_url: connection URI for Redis
    :param concurrency: number of tasks the worker can run simultaneously
    :param sync_concurrency:
        max number of synchronous tasks the worker can run simultaneously
        in separate threads; defaults to the same as ``concurrency``
    :param queue_name: name of queue in Redis
    :param queue_fetch_limit: max number of tasks to prefetch from Redis
    :param lifespan:
        async context manager that wraps worker execution and provides task
        dependencies
    :param serializer: function to serialize task data for Redis
    :param deserializer: function to deserialize task data from Redis
    :param tz: timezone to use for cron jobs
    :param handle_signals: whether to handle signals for graceful shutdown
    :param health_crontab: crontab for frequency to log and store health info
    :param with_scheduler:
        whether to run a scheduler alongside the worker; if None, the CLI will run
        a single scheduler instead of each worker running one. Multiple schedulers
        can be redundant, so running just one may slightly improve performance.
    """

    __slots__ = (
        "redis",
        "concurrency",
        "queue_name",
        "_group_name",
        "queue_fetch_limit",
        "bs",
        "counters",
        "loop",
        "_deps",
        "scripts",
        "registry",
        "cron_jobs",
        "cron_schedule",
        "id",
        "serializer",
        "deserializer",
        "task_wrappers",
        "tasks",
        "tz",
        "aborting_tasks",
        "burst",
        "_handle_signals",
        "_block_new_tasks",
        "_aentered",
        "_aexited",
        "lifespan",
        "_queue_key",
        "_stream_key",
        "_abort_key",
        "_health_key",
        "_channel_key",
        "_timeout_key",
        "main_task",
        "_start_time",
        "_prefix",
        "sync_concurrency",
        "_limiter",
        "_sentinel",
        "_health_tab",
        "with_scheduler",
        "middlewares",
    )

    def __init__(
        self,
        redis_url: str = "redis://localhost:6379",
        redis_sentinel_nodes: list[tuple[str, int]] | None = None,
        redis_sentinel_master: str = "mymaster",
        concurrency: int = 16,
        sync_concurrency: int | None = None,
        queue_name: str = DEFAULT_QUEUE_NAME,
        queue_fetch_limit: int | None = None,
        lifespan: Callable[["Worker"], AbstractAsyncContextManager[WD]] = _lifespan,
        serializer: Callable[[Any], EncodableT] = pickle.dumps,
        deserializer: Callable[[Any], Any] = pickle.loads,
        tz: tzinfo = timezone.utc,
        handle_signals: bool = True,
        health_crontab: str = "*/5 * * * *",
        with_scheduler: bool | None = None,
    ):
        #: Redis connection
        if redis_sentinel_nodes:
            self._sentinel = Sentinel(redis_sentinel_nodes, socket_timeout=2.0)
            self.redis = self._sentinel.master_for(redis_sentinel_master)
        else:
            self.redis = Redis.from_url(redis_url)
        self.concurrency = concurrency
        self.queue_name = queue_name
        self._group_name = REDIS_GROUP
        self.queue_fetch_limit = queue_fetch_limit or concurrency * 2
        #: semaphore controlling concurrency
        self.bs = asyncio.BoundedSemaphore(concurrency)
        #: mapping of type of task -> number of tasks of that type
        #: eg ``{"completed": 4, "failed": 1, "retried": 0}``
        self.counters = defaultdict(int)
        #: event loop for running tasks
        self.loop = asyncio.get_event_loop()
        self._deps = uninitialized
        #: Redis scripts for common operations
        self.scripts: dict[str, AsyncScript] = {}
        #: mapping of task name -> task wrapper
        self.registry: dict[str, RegisteredCron | RegisteredTask] = {}
        #: mapping of task name -> cron wrapper
        self.cron_jobs: dict[str, RegisteredCron] = {}
        #: mapping of task name -> next execution time in ms
        self.cron_schedule: dict[str, int] = defaultdict(int)
        #: unique ID of worker
        self.id = uuid4().hex[:8]
        self.serializer = serializer
        self.deserializer = deserializer
        #: mapping of task ID -> asyncio Task wrapper
        self.task_wrappers: dict[str, asyncio.Task] = {}
        #: mapping of task ID -> asyncio Task for task
        self.tasks: dict[str, asyncio.Task] = {}
        self._handle_signals = handle_signals
        self.tz = tz
        #: set of tasks currently scheduled for abortion
        self.aborting_tasks: set[str] = set()
        #: whether to shut down the worker when the queue is empty; set via CLI
        self.burst = False
        #: list of middlewares added to the worker
        self.middlewares: list[
            Callable[
                [WrappedContext[WD], Callable[..., Coroutine]], Callable[..., Coroutine]
            ]
        ] = []
        self.with_scheduler = with_scheduler
        self.sync_concurrency = sync_concurrency or concurrency
        self._limiter = CapacityLimiter(self.sync_concurrency)
        self._block_new_tasks = False
        self._aentered = False
        self._aexited = False
        self.lifespan = lifespan(self)
        self._prefix = REDIS_PREFIX + self.queue_name
        self._queue_key = self._prefix + REDIS_QUEUE
        self._stream_key = self._prefix + REDIS_STREAM
        self._abort_key = self._prefix + REDIS_ABORT
        self._health_key = self._prefix + REDIS_HEALTH
        self._channel_key = self._prefix + REDIS_CHANNEL
        self._timeout_key = self._prefix + REDIS_TIMEOUT
        self._start_time = now_ms()
        self._health_tab = CronTab(health_crontab)

        @self.cron(health_crontab, ttl=0)
        async def redis_health_check(ctx: WrappedContext[WD]) -> None:
            """
            Saves Redis health in Redis, then logs worker and Redis health.
            """
            async with self.redis.pipeline(transaction=True) as pipe:
                pipe.info(section="Memory")
                pipe.info(section="Clients")
                pipe.dbsize()
                for priority in TaskPriority:
                    pipe.xlen(self._stream_key + priority.value)
                pipe.zcard(self._queue_key)
                (
                    info_memory,
                    info_clients,
                    key_count,
                    low_size,
                    medium_size,
                    high_size,
                    queue_size,
                ) = await pipe.execute()

            mem_usage = info_memory.get("used_memory_human", "?")
            clients = info_clients.get("connected_clients", "?")
            queued = low_size + medium_size + high_size
            health = (
                f"redis {{memory: {mem_usage}, clients: {clients}, keys: {key_count}, "
                f"queued: {queued}, scheduled: {queue_size}}}"
            )
            async with self.redis.pipeline(transaction=True) as pipe:
                pipe.hset(self._health_key, "redis", health)
                pipe.hgetall(self._health_key)
                _, res = await pipe.execute()
            formatted = [v.decode() for v in res.values()]
            logger.info(
                "health check results:\n" + "\n".join(f for f in sorted(formatted))
            )

    @property
    def deps(self) -> WD:
        if self._deps == uninitialized:
            raise StreaqError(
                "Worker did not initialize correctly, are you using the async context manager?"
            )
        return cast(WD, self._deps)

    def build_context(
        self,
        fn_name: str,
        registered_task: RegisteredCron | RegisteredTask,
        id: str,
        tries: int = 1,
    ) -> WrappedContext[WD]:
        """
        Creates the context for a task to be run given task metadata
        """
        return WrappedContext(
            deps=self.deps,
            fn_name=fn_name,
            redis=self.redis,
            task_id=id,
            timeout=registered_task.timeout,
            tries=tries,
            ttl=registered_task.ttl,
            worker_id=self.id,
        )

    def cron(
        self,
        tab: str,
        max_tries: int | None = 3,
        timeout: timedelta | int | None = None,
        ttl: timedelta | int | None = timedelta(minutes=5),
        unique: bool = True,
    ):
        """
        Registers a task to be run at regular intervals as specified.

        :param tab:
            crontab for scheduling, follows the specification
            `here <https://github.com/josiahcarlson/parse-crontab?tab=readme-ov-file#description>`_.
        :param max_tries:
            number of times to retry the task should it fail during execution
        :param timeout: time after which to abort the task, if None will never time out
        :param ttl: time to store results in Redis, if None will never expire
        :param unique: whether multiple instances of the task can exist simultaneously
        """

        def wrapped(
            fn: Callable[[WrappedContext[WD]], Coroutine[Any, Any, R] | R],
        ) -> RegisteredCron[WD, R]:
            if not asyncio.iscoroutinefunction(fn):
                _fn: Callable[[WrappedContext[WD]], Coroutine[Any, Any, R]] = asyncify(
                    fn, self._limiter
                )  # type: ignore
            else:
                _fn = fn
            task = RegisteredCron(
                _fn,
                max_tries,
                CronTab(tab),
                timeout,
                ttl,
                unique,
                self,
            )
            self.cron_jobs[task.fn_name] = task
            self.registry[task.fn_name] = task
            logger.debug(f"cron job {task.fn_name} registered in worker {self.id}")
            return task

        return wrapped

    def task(
        self,
        max_tries: int | None = 3,
        timeout: timedelta | int | None = None,
        ttl: timedelta | int | None = timedelta(minutes=5),
        unique: bool = False,
    ):
        """
        Registers a task with the worker which can later be enqueued by the user.

        :param max_tries:
            number of times to retry the task should it fail during execution
        :param timeout: time after which to abort the task, if None will never time out
        :param ttl: time to store results in Redis, if None will never expire
        :param unique: whether multiple instances of the task can exist simultaneously
        """

        def wrapped(
            fn: Callable[
                Concatenate[WrappedContext[WD], P], Coroutine[Any, Any, R] | R
            ],
        ) -> RegisteredTask[WD, P, R]:
            if not asyncio.iscoroutinefunction(fn):
                _fn: Callable[
                    Concatenate[WrappedContext[WD], P], Coroutine[Any, Any, R]
                ] = asyncify(fn, self._limiter)  # type: ignore
            else:
                _fn = fn
            task = RegisteredTask(
                _fn,
                max_tries,
                timeout,
                ttl,
                unique,
                self,
            )
            self.registry[task.fn_name] = task
            logger.debug(f"task {task.fn_name} registered in worker {self.id}")
            return task

        return wrapped

    def middleware(
        self,
        fn: Callable[
            [WrappedContext[WD], Callable[..., Coroutine]], Callable[..., Coroutine]
        ],
    ):
        """
        Registers the given middleware with the worker.
        """
        self.middlewares.append(fn)
        return fn

    def run_sync(self) -> None:
        """
        Sync function to run the worker, finally closes worker connections.
        """
        self.main_task = self.loop.create_task(self.main())
        try:
            self.loop.run_until_complete(self.main_task)
        except asyncio.CancelledError:
            logger.debug(f"main loop interrupted, closing worker {self.id}")
        finally:
            self.loop.run_until_complete(self.close())

    async def run_async(self) -> None:
        """
        Async function to run the worker. Cleanup should be handled separately.
        """
        self.main_task = self.loop.create_task(self.main())
        await self.main_task

    async def main(self) -> None:
        """
        Main loop for handling worker tasks, aggregates and runs other tasks
        """
        logger.info(f"starting worker {self.id} for {len(self)} functions")
        # register signal handlers
        if self._handle_signals:
            self._add_signal_handler(signal.SIGINT)
            self._add_signal_handler(signal.SIGTERM)
        # create consumer group if it doesn't exist
        async with self.redis.pipeline(transaction=False) as pipe:
            for priority in TaskPriority:
                pipe.xgroup_create(
                    name=self._stream_key + priority.value,
                    groupname=self._group_name,
                    id=0,
                    mkstream=True,
                )
            await pipe.execute(raise_on_error=False)
            logger.debug(f"consumer group {self._group_name} created for queues")
        async with self:
            # run loops
            tasks = [self.listen_stream(), self.health_check()]
            if self.with_scheduler:
                tasks.append(self.listen_queue())
            futures = [self.loop.create_task(t) for t in tasks]
            try:
                _, pending = await asyncio.wait(
                    futures,
                    return_when=asyncio.FIRST_COMPLETED,
                )
                logger.debug(f"main loop wrapping up execution for worker {self.id}")
                for task in pending:
                    task.cancel()
            except asyncio.CancelledError:
                for task in futures:
                    task.cancel()
                await asyncio.gather(*futures, return_exceptions=True)

    async def listen_queue(self) -> None:
        """
        Periodically check the future queue (sorted set) for tasks, adding
        them to the live queue (stream) when ready, as well as adding cron
        jobs to the live queue when ready.
        """
        message_prefix = self._prefix + REDIS_MESSAGE
        # schedule initial cron jobs
        futures = set()
        for name, cron_job in self.cron_jobs.items():
            self.cron_schedule[name] = cron_job.next()
            futures.add(cron_job.enqueue().start(schedule=cron_job.schedule()))
        if futures:
            logger.debug(f"enqueuing {len(futures)} cron jobs in worker {self.id}")
            await asyncio.gather(*futures)
        while not self._block_new_tasks:
            start_time = time()
            task_ids = await self.redis.zrange(
                self._queue_key,
                start=0,
                end=now_ms(),
                num=self.queue_fetch_limit,
                offset=0,
                withscores=True,
                byscore=True,
            )
            async with self.redis.pipeline(transaction=False) as pipe:
                for priority in TaskPriority:
                    await self.scripts["unclaim_idle_tasks"](
                        keys=[
                            self._timeout_key + priority.value,
                            self._stream_key + priority.value,
                            self._group_name,
                            self.id,
                            message_prefix,
                        ],
                        args=[now_ms(), DEFAULT_TTL],
                        client=pipe,
                    )
                for task_id, score in task_ids:
                    expire_ms = int(score - now_ms() + DEFAULT_TTL)
                    if expire_ms <= 0:
                        expire_ms = DEFAULT_TTL

                    decoded = task_id.decode()
                    await self.scripts["publish_delayed_task"](
                        keys=[
                            self._queue_key,
                            self._stream_key,
                            self._prefix + REDIS_MESSAGE + decoded,
                        ],
                        args=[decoded, expire_ms, TaskPriority.MEDIUM.value],
                        client=pipe,
                    )
                if task_ids:
                    logger.debug(
                        f"enqueuing {len(task_ids)} delayed tasks in worker {self.id}"
                    )
                res = await pipe.execute()
            idle = res[0] + res[1] + res[2]  # for each priority level
            if idle:
                logger.info(f"retrying ↻ {len(idle)} idle tasks")

            # cron jobs
            futures = set()
            ts = now_ms()
            for name, cron_job in self.cron_jobs.items():
                if ts - 500 > self.cron_schedule[name]:
                    self.cron_schedule[name] = cron_job.next()
                    futures.add(cron_job.enqueue().start(schedule=cron_job.schedule()))
            if futures:
                logger.debug(f"enqueuing {len(futures)} cron jobs in worker {self.id}")
                await asyncio.gather(*futures)

            delay = time() - start_time
            if delay < 0.5:
                await asyncio.sleep(0.5 - delay)

    async def listen_stream(self) -> None:
        """
        Listen for new tasks from the stream, and periodically check for tasks
        that were never XACK'd but have timed out to reclaim.
        """
        high_stream = self._stream_key + TaskPriority.HIGH.value
        medium_stream = self._stream_key + TaskPriority.MEDIUM.value
        low_stream = self._stream_key + TaskPriority.LOW.value
        priority_order = {
            TaskPriority.HIGH.value: 0,
            TaskPriority.MEDIUM.value: 1,
            TaskPriority.LOW.value: 2,
        }
        while not self._block_new_tasks:
            messages: list[StreamMessage] = []
            active_tasks = self.concurrency - self.bs._value
            pending_tasks = len(self.task_wrappers)
            count = self.queue_fetch_limit - pending_tasks
            async with self.redis.pipeline(transaction=False) as pipe:
                pipe.smembers(self._abort_key)
                if count > 0:
                    pipe.xreadgroup(
                        groupname=self._group_name,
                        consumername=self.id,
                        streams={high_stream: ">", medium_stream: ">", low_stream: ">"},
                        block=500,
                        count=count,
                    )
                res = await pipe.execute()
            if count > 0:
                for stream, msgs in res[-1]:
                    priority = stream.decode().split(":")[-1]
                    messages.extend(
                        [
                            StreamMessage(
                                message_id=msg_id.decode(),
                                task_id=msg[b"task_id"].decode(),
                                priority=priority,
                            )
                            for msg_id, msg in msgs
                        ]
                    )
            else:
                await asyncio.sleep(0)  # yield control
            # Go through task_ids in the aborted tasks set and cancel those tasks.
            if res[0]:
                aborted: set[str] = set()
                for task_id_bytes in res[0]:
                    task_id = task_id_bytes.decode()
                    if task_id in self.tasks:
                        self.tasks[task_id].cancel()
                        aborted.add(task_id)
                if aborted:
                    logger.debug(f"aborting {len(aborted)} tasks in worker {self.id}")
                    self.aborting_tasks.update(aborted)
                    await self.redis.srem(self._abort_key, *aborted)  # type: ignore
            # start new tasks
            if messages:
                logger.debug(f"starting {len(messages)} new tasks in worker {self.id}")
                messages.sort(key=lambda msg: priority_order[msg.priority])
            for message in messages:
                coro = self.run_task(message)
                self.task_wrappers[message.task_id] = self.loop.create_task(coro)
            # wrap things up if we burstin'
            if (
                self.burst
                and not messages
                and active_tasks == 0
                and count > 0
                and not self.task_wrappers
            ):
                self._block_new_tasks = True
            # cleanup aborted tasks
            for task_id, task in list(self.task_wrappers.items()):
                if task.done():
                    del self.task_wrappers[task_id]
                    # propagate error
                    task.result()

    async def run_task(self, msg: StreamMessage) -> None:
        """
        Execute the registered task, then store the result in Redis.
        """
        task_id = msg.task_id
        key = lambda mid: self._prefix + mid + task_id
        # acquire semaphore
        async with self.bs:
            start_time = now_ms()
            async with self.redis.pipeline(transaction=True) as pipe:
                pipe.get(key(REDIS_TASK))
                pipe.incr(key(REDIS_RETRY))
                pipe.srem(self._abort_key, task_id)
                pipe.pexpire(key(REDIS_RETRY), DEFAULT_TTL)
                raw, task_try, abort, _ = await pipe.execute()

            async def handle_failure(
                exc: BaseException, ttl: timedelta | int | None = 300
            ) -> None:
                self.counters["failed"] += 1
                data = {
                    "s": False,
                    "r": exc,
                    "st": start_time,
                    "ft": now_ms(),
                }
                try:
                    raw = self.serializer(data)
                    await asyncio.shield(self.finish_failed_task(msg, raw, ttl))
                except Exception as e:
                    raise StreaqError(
                        f"Failed to serialize result for task {task_id}!"
                    ) from e

            if not raw:
                logger.warning(f"task {task_id} expired †")
                return await handle_failure(StreaqError("Task execution failed!"))

            try:
                data = self.deserializer(raw)
            except Exception as e:
                logger.exception(f"Failed to deserialize task {task_id}: {e}")
                return await handle_failure(e)

            fn_name = data["f"]
            if fn_name not in self.registry:
                logger.error(
                    f"Missing function {fn_name}, can't execute task {task_id}!"
                )
                return await handle_failure(StreaqError("Nonexistent function!"))
            task = self.registry[fn_name]

            if abort:
                logger.info(f"task {task_id} aborted ⊘ prior to run")
                return await handle_failure(asyncio.CancelledError(), ttl=task.ttl)
            if task.max_tries and task_try > task.max_tries:
                logger.warning(
                    f"task {task_id} failed × after {task.max_tries} retries"
                )
                return await handle_failure(
                    StreaqError(f"Max retry attempts reached for task {task_id}!"),
                    ttl=task.ttl,
                )
            timeout = (
                None
                if task.timeout is None
                else start_time + 1000 + to_ms(task.timeout)
            )
            after = data.get("A")
            async with self.redis.pipeline(transaction=True) as pipe:
                pipe.set(key(REDIS_RUNNING), 1, pxat=timeout)
                if timeout:
                    pipe.zadd(
                        self._timeout_key + msg.priority, {msg.message_id: timeout}
                    )
                if after:
                    pipe.get(self._prefix + REDIS_RESULT + after + ":raw")
                res = await pipe.execute()
            _args = data["a"] if not after else self.deserializer(res[-1])

            ctx = self.build_context(fn_name, task, task_id, tries=task_try)
            success = True
            delay = None
            done = True
            finish_time = None

            async def _fn(ctx, *args, **kwargs):
                return await asyncio.wait_for(
                    task.fn(ctx, *args, **kwargs),
                    to_seconds(task.timeout) if task.timeout is not None else None,
                )

            try:
                logger.info(f"task {task_id} → worker {self.id}")
                wrapped = _fn
                for middleware in reversed(self.middlewares):
                    wrapped = middleware(ctx, wrapped)
                coro = wrapped(ctx, *_args, **data["k"])
                try:
                    self.tasks[task_id] = self.loop.create_task(coro)
                    result = await self.tasks[task_id]
                except Exception as e:
                    raise e  # re-raise for outer try/except
                finally:
                    del self.tasks[task_id]
                    finish_time = now_ms()
            except StreaqRetry as e:
                result = e
                success = False
                done = False
                if e.delay is not None:
                    delay = to_seconds(e.delay)
                else:
                    delay = task_try**2
                logger.exception(e)
                logger.info(f"retrying ↻ task {task_id} in {delay}s")
            except asyncio.TimeoutError as e:
                logger.error(f"task {task_id} timed out …")
                result = e
                success = False
                done = True
            except asyncio.CancelledError as e:
                if task_id in self.aborting_tasks:
                    logger.info(f"task {task_id} aborted ⊘")
                    self.aborting_tasks.remove(task_id)
                    done = True
                    self.counters["aborted"] += 1
                    self.counters["failed"] -= 1  # this will get incremented later
                else:
                    logger.info(f"task {task_id} cancelled, will be retried ↻")
                    done = False
                result = e
                success = False
            except Exception as e:
                result = e
                success = False
                done = True
                logger.info(f"task {task_id} failed ×")
                logger.exception(e)
            finally:
                await asyncio.shield(
                    self.finish_task(
                        msg,
                        finish=done,
                        delay=delay,
                        return_value=result,  # type: ignore
                        start_time=start_time,
                        finish_time=finish_time or now_ms(),
                        success=success,
                        ttl=task.ttl,
                        triggers=data.get("T"),
                    )
                )

    async def finish_task(
        self,
        msg: StreamMessage,
        finish: bool,
        delay: float | None,
        return_value: Any,
        start_time: int,
        finish_time: int,
        success: bool,
        ttl: timedelta | int | None,
        triggers: str | None,
    ) -> None:
        """
        Cleanup for a task that executed successfully or will be retried.
        """
        data = {
            "s": success,
            "r": return_value,
            "st": start_time,
            "ft": finish_time,
        }
        task_id = msg.task_id
        try:
            result = self.serializer(data)
        except Exception as e:
            raise StreaqError(f"Failed to serialize result for task {task_id}!") from e
        key = lambda mid: self._prefix + mid + task_id
        stream_key = self._stream_key + msg.priority
        async with self.redis.pipeline(transaction=True) as pipe:
            pipe.xack(stream_key, self._group_name, msg.message_id)
            pipe.xdel(stream_key, msg.message_id)
            pipe.zrem(self._timeout_key + msg.priority, msg.message_id)
            if finish:
                pipe.publish(self._channel_key, task_id)
                if success:
                    self.counters["completed"] += 1
                else:
                    self.counters["failed"] += 1
                if result and ttl != 0:
                    pipe.set(key(REDIS_RESULT), result, ex=ttl)
                pipe.delete(
                    key(REDIS_RETRY),
                    key(REDIS_RUNNING),
                    key(REDIS_TASK),
                    key(REDIS_MESSAGE),
                )
                pipe.srem(self._abort_key, task_id)
                if success:
                    ret, trunc = str(return_value), 32
                    if len(ret) > trunc:
                        ret = f"{ret[:trunc]}…"
                    logger.info(f"task {task_id} ← {ret}")
                    if triggers:
                        args = self.serializer(to_tuple(return_value))
                        pipe.set(
                            key(REDIS_RESULT) + ":raw", args, ex=timedelta(minutes=5)
                        )
                    script = self.scripts["update_dependents"]
                else:
                    script = self.scripts["fail_dependents"]
                await script(
                    keys=[
                        self._prefix + REDIS_DEPENDENTS,
                        self._prefix + REDIS_DEPENDENCIES,
                        task_id,
                    ],
                    client=pipe,
                )
            elif delay:
                self.counters["retried"] += 1
                pipe.delete(key(REDIS_MESSAGE))
                pipe.zadd(self._queue_key, {task_id: now_ms() + delay * 1000})
            else:
                self.counters["retried"] += 1
                ttl_ms = to_ms(ttl) if ttl is not None else None
                expire = (ttl_ms or 0) + DEFAULT_TTL
                await self.scripts["retry_task"](
                    keys=[stream_key, key(REDIS_MESSAGE)],
                    args=[task_id, expire],
                    client=pipe,
                )
            res = await pipe.execute()
        if finish and res[-1]:
            if success:
                async with self.redis.pipeline(transaction=False) as pipe:
                    for dep_id in res[-1]:
                        dep_id = dep_id.decode()
                        await self.scripts["publish_dependent"](
                            keys=[
                                stream_key,
                                self._prefix + REDIS_MESSAGE + dep_id,
                                dep_id,
                            ],
                            args=[DEFAULT_TTL],
                            client=pipe,
                        )
                    await pipe.execute()
            else:
                failure = {
                    "s": False,
                    "r": StreaqError("Dependency failed, not running task!"),
                    "st": -1,
                    "ft": -1,
                }
                try:
                    result = self.serializer(failure)
                except Exception as e:
                    raise StreaqError(
                        f"Failed to serialize result for dependency of task {task_id}!"
                    ) from e
                async with self.redis.pipeline(transaction=False) as pipe:
                    self.counters["failed"] += len(res[-1])
                    for dep_id in res[-1]:
                        dep_id = dep_id.decode()
                        logger.info(f"task {dep_id} dependency failed ×")
                        await self.scripts["unpublish_dependent"](
                            keys=[
                                self._prefix + REDIS_TASK + dep_id,
                                self._prefix + REDIS_RESULT + dep_id,
                                self._channel_key,
                                dep_id,
                            ],
                            args=[result],
                            client=pipe,
                        )
                    await pipe.execute()

    async def finish_failed_task(
        self,
        msg: StreamMessage,
        result_data: EncodableT,
        ttl: timedelta | int | None,
    ) -> None:
        """
        Cleanup for a task that failed during execution.
        """
        task_id = msg.task_id
        key = lambda mid: self._prefix + mid + task_id
        self.counters["failed"] += 1
        stream_key = self._stream_key + msg.priority
        async with self.redis.pipeline(transaction=True) as pipe:
            pipe.delete(
                key(REDIS_RETRY),
                key(REDIS_RUNNING),
                key(REDIS_TASK),
                key(REDIS_MESSAGE),
            )
            pipe.publish(self._channel_key, task_id)
            pipe.srem(self._abort_key, task_id)
            pipe.xack(stream_key, self._group_name, msg.message_id)
            pipe.xdel(stream_key, msg.message_id)
            pipe.zrem(self._timeout_key + msg.priority, msg.message_id)
            if result_data is not None:
                pipe.set(key(REDIS_RESULT), result_data, ex=ttl)
            await self.scripts["fail_dependents"](
                keys=[
                    self._prefix + REDIS_DEPENDENTS,
                    self._prefix + REDIS_DEPENDENCIES,
                    task_id,
                ],
                client=pipe,
            )
            res = await pipe.execute()
        if res[-1]:
            failure = {
                "s": False,
                "r": StreaqError("Dependency failed, not running task!"),
                "st": -1,
                "ft": -1,
            }
            try:
                result = self.serializer(failure)
            except Exception as e:
                raise StreaqError(
                    f"Failed to serialize result for task {task_id}!"
                ) from e
            async with self.redis.pipeline(transaction=False) as pipe:
                self.counters["failed"] += len(res[-1])
                for dep_id in res[-1]:
                    dep_id = dep_id.decode()
                    logger.info(f"task {dep_id} dependency failed ×")
                    await self.scripts["unpublish_dependent"](
                        keys=[
                            self._prefix + REDIS_TASK + dep_id,
                            self._prefix + REDIS_RESULT + dep_id,
                            self._channel_key,
                            dep_id,
                        ],
                        args=[result],
                        client=pipe,
                    )
                await pipe.execute()

    def enqueue_unsafe(
        self,
        fn_name: str,
        *args,
        unique: bool = False,
        **kwargs,
    ) -> Task:
        """
        Allows for enqueuing a task that is registered elsewhere without having access
        to the worker it's registered to. This is unsafe because it doesn't check if the
        task is registered with the worker and doesn't enforce types, so it should only
        be used if you need to separate the task queuing and task execution code for
        performance reasons.

        :param fn_name:
            name of the function to run, much match its __qualname__. If you're unsure,
            check ``Worker.registry``.
        :param unique: whether multiple instances of the task can exist simultaneously
        :param args: positional arguments for the task
        :param kwargs: keyword arguments for the task

        :return: task object
        """
        registered = RegisteredTask(
            fn=_placeholder,
            max_tries=None,
            timeout=None,
            ttl=None,
            unique=unique,
            worker=self,
            _fn_name=fn_name,
        )
        return Task(args, kwargs, registered)

    async def queue_size(self) -> int:
        """
        Returns the number of tasks currently queued in Redis.
        """
        async with self.redis.pipeline(transaction=True) as pipe:
            for priority in TaskPriority:
                pipe.xlen(self._stream_key + priority.value)
            pipe.zcard(self._queue_key)
            res = await pipe.execute()

        return sum(res)

    @property
    def active(self) -> int:
        """
        The number of currently active tasks for the worker
        """
        return sum(not t.done() for t in self.tasks.values())

    async def health_check(self) -> None:
        """
        Periodically stores info about the worker in Redis.
        """
        while True:
            await asyncio.sleep(self._health_tab.next(now=datetime.now(self.tz)))  # type: ignore
            async with self.redis.pipeline(transaction=True) as pipe:
                pipe.hset(self._health_key, self.id, str(self))
                await pipe.execute()

    def _add_signal_handler(self, signum: Signals) -> None:
        try:
            self.loop.add_signal_handler(signum, partial(self.handle_signal, signum))
        except NotImplementedError:
            logger.error("Windows does not support handling Unix signals!")

    def handle_signal(self, signum: Signals) -> None:
        """
        Gracefully shutdown the worker when a signal is received.
        """
        logger.info(f"received signal {signum.name}, shutting down worker {self.id}")
        self._block_new_tasks = True

    async def close(self) -> None:
        """
        Cleanup worker and Redis connection
        """
        self._block_new_tasks = True
        for t in self.task_wrappers.values():
            if not t.done():
                t.cancel()
        self.main_task.cancel()
        await asyncio.gather(
            *self.task_wrappers.values(), self.main_task, return_exceptions=True
        )
        # delete consumers
        async with self.redis.pipeline(transaction=False) as pipe:
            for priority in TaskPriority:
                pipe.xgroup_delconsumer(
                    name=self._stream_key + priority.value,
                    groupname=self._group_name,
                    consumername=self.id,
                )
            pipe.hdel(self._health_key, self.id)
            await pipe.execute(raise_on_error=False)
        await self.redis.close(close_connection_pool=True)
        run_time = now_ms() - self._start_time
        logger.info(f"shutdown {str(self)} after {run_time}ms")

    async def _listen_for_result(self, pubsub: PubSub, task_id: str) -> None:
        encoded_id = task_id.encode()
        async for msg in pubsub.listen():
            if msg.get("data") == encoded_id:
                break

    async def status_by_id(self, task_id: str) -> TaskStatus:
        """
        Fetch the current status of the given task. Note that nonexistent
        tasks will return pending.

        :param task_id: ID of the task to check

        :return: status of the task
        """
        key = lambda mid: self._prefix + mid + task_id
        async with self.redis.pipeline(transaction=True) as pipe:
            pipe.exists(key(REDIS_RESULT))
            pipe.exists(key(REDIS_RUNNING))
            pipe.zscore(self._queue_key, task_id)
            pipe.exists(key(REDIS_MESSAGE))
            finished, running, score, queued = await pipe.execute()

        if finished:
            return TaskStatus.DONE
        elif running:
            return TaskStatus.RUNNING
        elif score:
            return TaskStatus.SCHEDULED
        elif queued:
            return TaskStatus.QUEUED
        return TaskStatus.PENDING

    async def result_by_id(
        self, task_id: str, timeout: timedelta | int | None = None
    ) -> TaskResult[Any]:
        """
        Wait for and return the given task's result, optionally with a timeout.

        :param task_id: ID of the task to get results for
        :param timeout: amount of time to wait before raising a `TimeoutError`

        :return: wrapped result object
        """
        pubsub = self.redis.pubsub()
        await pubsub.subscribe(self._channel_key)
        key = lambda mid: self._prefix + mid + task_id
        raw = await self.redis.get(key(REDIS_RESULT))
        if raw is None:
            timeout_seconds = to_seconds(timeout) if timeout is not None else None
            await asyncio.wait_for(
                self._listen_for_result(pubsub, task_id), timeout_seconds
            )
            raw = await self.redis.get(key(REDIS_RESULT))
        if raw is None:
            raise StreaqError(
                "Task finished but result was not stored, did you set ttl=0?"
            )
        try:
            data = self.deserializer(raw)
            return TaskResult(
                success=data["s"],
                result=data["r"],
                start_time=data["st"],
                finish_time=data["ft"],
                queue_name=self.queue_name,
            )
        except Exception as e:
            raise StreaqError(
                f"Unable to deserialize result for task {task_id}:"
            ) from e

    async def abort_by_id(self, task_id: str, timeout: timedelta | int = 5) -> bool:
        """
        Notify workers that the task should be aborted.

        :param task_id: ID of the task to abort
        :param timeout: how long to wait to confirm abortion was successful

        :return: whether the task was aborted successfully
        """
        await self.redis.sadd(self._abort_key, task_id)  # type: ignore
        try:
            result = await self.result_by_id(task_id, timeout=timeout)
            return not result.success and isinstance(
                result.result, asyncio.CancelledError
            )
        except asyncio.TimeoutError:
            return False

    async def info_by_id(self, task_id: str) -> TaskData:
        """
        Fetch info about a previously enqueued task.

        :param task_id: ID of the task to get info for

        :return: task info object
        """
        key = lambda mid: self._prefix + mid + task_id
        async with self.redis.pipeline(transaction=False) as pipe:
            pipe.get(key(REDIS_TASK))
            pipe.get(key(REDIS_RETRY))
            pipe.zscore(self._queue_key, task_id)
            raw, task_try, score = await pipe.execute()
        data = self.deserializer(raw)
        dt = datetime.fromtimestamp(score / 1000, tz=self.tz) if score else None
        return TaskData(
            fn_name=data["f"],
            enqueue_time=data["t"],
            task_try=task_try,
            scheduled=dt,
        )

    async def __aenter__(self):
        # reentrant
        if not self._aentered:
            self._aentered = True
            # register lua scripts
            self.scripts.update(register_scripts(self.redis))
            # user-defined deps
            self._deps = await self.lifespan.__aenter__()
        return self

    async def __aexit__(self, *exc):
        # reentrant
        if not self._aexited:
            self._aexited = True
            await self.lifespan.__aexit__(*exc)

    def __len__(self):
        return len(self.registry)

    def __str__(self) -> str:
        counters_str = dict.__repr__(self.counters).replace("'", "")
        return f"worker {self.id} {counters_str}"

    def __repr__(self) -> str:
        return f"<{str(self)}>"
